function [gains, pow] = fbCntrlDes(f,Zi,H,S,motorSpecs,plotflag)
% [gains, pow] = fbCntrlDes(f,Zi,H,S,motorSpecs,plotflag)
%
% Finds optimal PI controller gains (designed for WaveBot).
%
% Inputs
%   f           frequency vector [Hz]
%   Zi          wave-body impedance FRF
%   H           excitation FRF
%   S           wave spectrum (same formatting as WAFO)
%   motorSpecs  structure with motor parameters (see below)
%   plotflag    1 to produce plots
%
% Outputs
%   gains       [Kp, Ki]
%   pow         power [w]
%
% Copyright 2019 NTESS
%
% Licensed under the Apache License, Version 2.0 (the "License");
% you may not use this file except in compliance with the License.
% You may obtain a copy of the License at
%
%     http://www.apache.org/licenses/LICENSE-2.0
%
% Unless required by applicable law or agreed to in writing, software
% distributed under the License is distributed on an "AS IS" BASIS,
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
% See the License for the specific language governing permissions and
% limitations under the License.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%

if nargin < 5
    plotflag = 0;
end

if nargin < 4 || isempty(motorSpecs)
    Kt = 6.1745;    % WaveBot motor torque constant
    R = 0.5;        % WaveBot motor electrical winding resistance
    N = 12.4666;    % WaveBot heave gear ratio
else
    Kt = motorSpecs.Kt;
    R = motorSpecs.R;
    N = motorSpecs.N;
end

w = 2*pi*f;
n = length(f);

%% Wave spectrum and excitation

Hm0 = spec2char(S,1);
Tp = spec2char(S,11);
Te = spec2char(S,5);

% amplitude spectrum
dw = S.w(2) - S.w(1);
ampSpect = sqrt(2*S.S*dw) .* exp(1i*2*pi*rand(n,1));

% complex excitation spectrum
Fe = H .* ampSpect;

%% Controllers

% define PI feedback controller FRF
piControlFrf = @(Kp,Ki) Kp - 1i*Ki./w;

% define P feedback (resistive damping) controller FRF
pControlFrf = @(Kp,Ki) Kp;

%% Objective functions

% define objective functions (power will be negative)
powObj_pi_e = @(x) WaveBot_fbPow(Zi,Fe,piControlFrf(x(1),x(2)),Kt,R,N);
powObj_pi_m = @(x) WaveBot_fbPow(Zi,Fe,piControlFrf(x(1),x(2)),Kt,0,N);
powObj_p_e = @(x) WaveBot_fbPow(Zi,Fe,pControlFrf(x(1),x(2)),Kt,R,N);

%% Optimization solution

x0 = [0,0];

opts = optimoptions('fminunc');
opts.Display = 'off';
[x_pi_e,fval_pi_e] = fminunc(powObj_pi_e,x0,opts);
[x_pi_m,fval_pi_m] = fminunc(powObj_pi_m,x0,opts);
[x_p_e,fval_p_e] = fminunc(powObj_p_e,x0,opts);

%% frequency dependent power

% PI controller tuned for electrical power electrical power
[~,pow_pi_e_f,pow_ub_f] = WaveBot_fbPow(Zi,Fe,piControlFrf(x_pi_e(1),x_pi_e(2)),Kt,R,N);

% PI controller tuned for electrical power mechanical power
[~,pow_pi_em_f,~] = WaveBot_fbPow(Zi,Fe,piControlFrf(x_pi_e(1),x_pi_e(2)),Kt,0,N);

% PI controller tuned for mechanical power mechanical power 
[~,pow_pi_m_f,~] = WaveBot_fbPow(Zi,Fe,piControlFrf(x_pi_m(1),x_pi_m(2)),Kt,0,N);

% PI contorller tuned for mechanical power electrical power
[fval_pi_me,pow_me_f,~] = WaveBot_fbPow(Zi,Fe,piControlFrf(x_pi_m(1),x_pi_m(2)),Kt,R,N);

% P controller tuned for electrical power electrical
[~,pow_p_e_f,~] = WaveBot_fbPow(Zi,Fe,pControlFrf(x_p_e(1),x_p_e(2)),Kt,R,N);

gains = [x_pi_e(1), x_pi_e(2)];
pow = fval_pi_e;

fprintf('\n----------------------------------------------------\n')
fprintf('Sig. wave height [m]:\t\t%.2f\n',Hm0)
fprintf('Peak period [s]:\t\t%.2f\n',Tp)
fprintf('Energy frequency [Hz]:\t\t%.3f\n',1/Te)
fprintf('Optimal gains (Kp, Ki):\t\t(%.1e, %.1e)\n',x_pi_e(1),x_pi_e(2))
fprintf('----------------------------------------------------\n')
fprintf('Theoretical power limit [W]:\t%.2e\n',sum(pow_ub_f))
fprintf('Optimal avg. power, elec. [W]:\t%.2e\n',fval_pi_e)
fprintf('\tPercent capture:\t%.1f%%\n',sum(pow_pi_e_f)/sum(pow_ub_f)*100)
fprintf('Optimal avg. power, mech. [W]:\t%.2e\n',fval_pi_m)
fprintf('\tPercent capture:\t%.1f%%\n\n',sum(pow_pi_m_f)/sum(pow_ub_f)*100)
fprintf('Avg. mech power, elec. optimal gains. [W]:\t%.2e\n',sum(pow_pi_em_f))
%% Plotting

if plotflag
    
    
    %------------------------------------
    figure('name','Zi Bode')
    
    ax(1) = subplot(2,1,1);
    semilogx(f,mag2db(abs(Zi)))
    hold on
    grid on
    ylabel('$\left\| Z_i \right\|$ [dB]','interpreter','latex')
    
    ax(2) = subplot(2,1,2);
    semilogx(f,unwrap(angle(Zi)))
    hold on
    grid on
    ylabel('$\angle Z_i$ [rad]','interpreter','latex')
    
    %------------------------------------
    figure('name','Zi real & imag.')
    
    ax(1) = subplot(2,1,1);
    semilogx(f,real(Zi))
    hold on
    grid on
    ylabel('Re$\left\{ Z_i \right\}$','interpreter','latex')
    
    ax(2) = subplot(2,1,2);
    semilogx(f,imag(Zi))
    hold on
    grid on
    ylabel('Im$\left\{ Z_i \right\}$','interpreter','latex')
    
    linkaxes(ax,'x')
    clear ax
    
    %------------------------------------
    figure('name','H real & imag.')
    
    ax(1) = subplot(2,1,1);
    semilogx(f,real(H))
    hold on
    grid on
    ylabel('Re$\left\{ H \right\}$','interpreter','latex')
    
    ax(2) = subplot(2,1,2);
    semilogx(f,imag(H))
    hold on
    grid on
    ylabel('Im$\left\{ H \right\}$','interpreter','latex')
    
    linkaxes(ax,'x')
    clear ax
    
    %------------------------------------
    figure('name','H Bode')
    
    ax(1) = subplot(2,1,1);
    semilogx(f,mag2db(abs(H)))
    hold on
    grid on
    ylabel('$\left\| H \right\|$ [dB]','interpreter','latex')
    
    ax(2) = subplot(2,1,2);
    semilogx(f,unwrap(angle(H)))
    hold on
    grid on
    ylabel('$\angle H$ [rad]','interpreter','latex')
    
    xlabel('Freq. [Hz]','interpreter','latex')
    
    linkaxes(ax,'x')
    
    %------------------------------------
    figure('name','Wave spectrum')
    hold on
    grid on
    plot(f,S.S*2*pi)
    ylabel('Spect. density [m$^2$/Hz]','interpreter','latex')
    xlabel('Freq. [Hz]','interpreter','latex')
    
    
    
    %------------------------------------
    Kps = linspace(x_pi_e(1)*2.5,0,20);
    Kis = linspace(-1.25*abs(x_pi_m(2)),1.25*abs(x_pi_m(2)),20);
    
    for ii = 1:length(Kps)
        for jj = 1:length(Kis)
            powm(ii,jj) = powObj_pi_e([Kps(ii),Kis(jj)]);
        end
    end
    
    [mgKp, mgKi] = meshgrid(Kps,Kis);
    
    pn = powm'./pow;
    minds = pn>0;
    mmask = nan(size(pn));
    mmask(minds) = 1;
    
    figure('name','Optimal power contour')
    title('$P^{elec}/P^{elec}_{max}$','interpreter','latex')
    hold on
    grid on
    ct = contourf(Kps,Kis,pn.*mmask,20);
    pl(1) = plot(x_pi_e(1),x_pi_e(2),'k*');
    pl(2) = plot(x_pi_m(1),x_pi_m(2),'k+');
    lg = legend(pl,...
        sprintf('Electrical optimal, $P^{elec}/P^{elec}_{max}=%.2f$',fval_pi_e/fval_pi_e),...
        sprintf('Mechanical optimal, $P^{elec}/P^{elec}_{max}=%.2f$',fval_pi_me/fval_pi_e));
    set(lg,'interpreter','latex')
    set(lg,'location','southeast')
    colormap('cool')
    cb = colorbar();
    caxis([0,1])
%     zlim([0,1])
    xlabel('$K_P$','interpreter','latex')
    ylabel('$K_I$','interpreter','latex')
    
    %------------------------------------
    Sn = S.S/S.w;
    Sn = Sn/max(Sn);
    Fexc=abs(Fe)./max(abs(Fe));
    pow_e_fn = pow_pi_e_f / min(pow_ub_f);
    pow_m_fn = pow_pi_m_f / min(pow_ub_f);
    pow_p_e_fn = pow_p_e_f / min(pow_ub_f);
    pow_ub_fn = pow_ub_f / min(pow_ub_f);
    pow_e_fn(isnan(pow_e_fn)) = 0;
    pow_m_fn(isnan(pow_m_fn)) = 0;
    figure('name','Capture efficiency')
    hold on
    grid on
    plot(f, Sn,'k--')
    plot(f, Fexc, 'k-.')
    plot(f,pow_ub_fn)
    plot(f, pow_m_fn)
    plot(f, pow_e_fn)
    plot(f, pow_p_e_fn)
    
    
    l1 = legend('$J / J_{max}$',...
        '$F_{exc}/F_{exc,\,max}$',...
        '$P^{mech}_{CC}/P^{mech,\,max}_{CC}$',...
        '$P^{mech}_{PI}/P^{mech,\,max}_{CC}$',...
        '$P^{elec}_{PI}/P^{mech,\,max}_{CC}$',...
        '$P^{elec}_{P}/P^{mech,\,max}_{CC}$');
    set(l1,'interpreter','latex')
    set(l1,'fontsize',12)
    xlabel('Freq. [Hz]','interpreter','latex')
    
end
