% Copyright 2019 NTESS
%
% Licensed under the Apache License, Version 2.0 (the "License");
% you may not use this file except in compliance with the License.
% You may obtain a copy of the License at
%
%     http://www.apache.org/licenses/LICENSE-2.0
%
% Unless required by applicable law or agreed to in writing, software
% distributed under the License is distributed on an "AS IS" BASIS,
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
% See the License for the specific language governing permissions and
% limitations under the License.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


clc
clear
close all

%% User input

IDs = 1:13;
Tps = [1.5800, 1.5800, 2.5000, 2.5000, 2.5000, 2.5000, 3.5000, 3.5000,...
    3.5000, 3.5000, 3.5000, 2.5000, 1.5800];
Hss = [0.127, 0.127, 0.127, 0.127, 0.254, 0.254, 0.127, 0.127, 0.254,...
    0.254, 0.381, 0.381, 0.254];
gammas = [1, 3.3, 1, 3.3, 1, 3.3, 1, 3.3, 1, 3.3, 3.3, 3.3, 3.3];

fprintf('Sea state')
fprintf('\n---------------------------------------------------------------------------------------------------')
fprintf('\nID:\t')
fprintf('%7i',IDs)
fprintf('\n---------------------------------------------------------------------------------------------------')
fprintf('\nHs [m]:\t')
fprintf('%7.3f',Hss)
fprintf('\nTp [s]:\t')
fprintf('%7.1f',Tps)
fprintf('\ngamma:\t')
fprintf('%7.1f',gammas)
fprintf('\n---------------------------------------------------------------------------------------------------')
fprintf('\n')
pID = input('Sea state ID: ');
ind = find(pID == IDs);
if isempty(ind)
    error('Sea state selection not found')
end
Hm0 = Hss(ind);
Tp = Tps(ind);
gamma = gammas(ind);

fprintf('\nMPC design parameters\n')
fprintf('---------------------\n')
predHor = input('Prediction horizon (default: 2) [s]: ');
Umax = input('PTO force limit (default: 1e10) [N]: ');
Zmax = input('Position limit (default: 1e10) [m]: ');
Rmax = input('PTO force slew rate limit (default: 1e10) [N/s]: ');
ITERmax = input('Max solver iterations (default: no limit): ');

%% Initial set up

% Load WEC model
mf = matfile('WaveBot_heaveModel.mat');
f = mf.f;               % frequency vector
H_frf = mf.H_frf;       % wave-body impedance FRF
Zi_frf = mf.Zi_frf;     % wave excitation FRF
df=mean(diff(f));       % model frequency spacing
f_0 = [(0:df:(min(f)-df)).' ; f];   % populate to zero frequency
H_0=[zeros(length(f_0)-length(f),1) ; H_frf]; % zero-pad wave-body impedance (no wave excitation here anyway)

% Create wave spectrum
Spect = jonswap(2*pi*f,[Hm0, Tp, gamma]);
T_rep = 50;            % repeat period of wave to generate(s)
dt = 0.01;              % the desired time step (s)
df = 1/T_rep;           % implied frequency domain spacing
f_vec(:,1) = 0 : df : 1; % frequency domain vector to create

%% PI tuning

[gains,pow] = fbCntrlDes(f,Zi_frf,H_frf,Spect,[],1);
Kp = gains(1);
Ki = gains(2);

%% MPC Design

if isempty(Umax)
    Umax = 1e10;
end
if isempty(Zmax)
    Zmax = 1e10;
end
if isempty(Rmax)
    Rmax = 1e10;
end
if isempty(predHor)
    predHor = 2;
end

[MPCobj,Plant,Sf] = mpcCntrlDes(f,Zi_frf,H_frf,...
    Spect,Kp,Ki,Umax,Zmax,Rmax,predHor,ITERmax);

%% Create Wave time series

% calculate excitation force time series
Spect_0 = jonswap(2*pi*f_vec,[Hm0, Tp, gamma]);                             % recalculate spectra from zero frequency
dw = Spect_0.w(2) - Spect_0.w(1);                                           % radian frequency spacing
ampSpect = sqrt(2*Spect_0.S*dw);                                            % calculate wave amplitude spectra from PSD
exc_mag=interp1(f_0,H_0,f_vec).*ampSpect;                                   % magnitude of excitation spectra                                 
rng(3);                                                                     % seeds rand for repeatable output;
phaseSeed=2*pi*rand(length(f_vec),1);                                       % randomizes phase of each freq component
exc_comp=exc_mag .* exp(1i.* phaseSeed);                                    % complex-value representation
exc_TS=ifft(exc_comp,T_rep/dt,'symmetric').*(T_rep/dt/2);                   % excitation time-series
t_vec=[0:dt:T_rep - dt];                                                    % excitation time vector (s)

% create structure for simulink input
exc_TS_sim=struct();
exc_TS_sim.time=t_vec;
exc_TS_sim.signals.values=exc_TS;
exc_TS_sim.signals.dimensions=[1];


%% Simulate w/ Simulink

open('fbCntrlDes_model')
VSS_NONE = Simulink.Variant('VSS_MODE==0');
VSS_PI = Simulink.Variant('VSS_MODE==1');
VSS_MPC = Simulink.Variant('VSS_MODE==2'); 

x0=zeros(1,2);
r=zeros(1,4);

for ii = [1,2]
    VSS_MODE = ii;
    simout{ii} = sim('fbCntrlDes_model',...
        'StopTime', num2str(T_rep-dt),...
        'ReturnWorkspaceOutputs','on');
end

%% Plot simulink result

figure('name','Simulink results')
N = simout{1}.logsout.numElements;
ax = [];
for ii = 1:N
    ax(ii) = subplot(N,1,ii);
    hold on
    grid on
    for jj = 1:length(simout)
        plot(simout{jj}.logsout.getElement(ii).Values)
    end
    ylabel(simout{1}.logsout.getElement(ii).Name)
    title('')
    if ii ~=N
        set(ax(ii),'XTickLabel',[])
        xlabel('')
    end
end
xlabel('Time [s]')
linkaxes(ax,'x')
xlim([0, simout{1}.SimulationMetadata.ModelInfo.StopTime])
l1 = legend('PI','MPC');
set(l1,'location','southwest')

%% Calculate simulink average powers

pow_mech_simPI= mean(simout{1}.logsout.getElement(5).Values);
fprintf('\n Simulation time-average mechanical power, PI [W]: \t%.2e\n', pow_mech_simPI);
pow_mech_simMPC= mean(simout{2}.logsout.getElement(5).Values);
fprintf('\n Simulation time-average mechanical power, MPC [W]: \t%.2e\n', pow_mech_simMPC);
