function [MPCobj,Plant,Sf] = mpcCntrlDes(f,Zi_frf,H_frf,Spect,Kp,Ki,Umax,Zmax,Rmax,predHoriz,maxIter)
% [MPCobj,Plant,Sf] = mpcCntrlDes(f,Zi_frf,H_frf,Spect,Kp,Ki,Umax,Zmax,Rmax,predHoriz,maxIter)
%
% Creates a feedback tuned MPC.
%
% Inputs:
%   f           frequency vector
%   Zi_frf      impedance freq. response function
%   H_frf       excitation freq. response function
%   Spect       spectrum (WAFO format)
%   Kp          proportional gain
%   Ki          integral gain
%   Umax        max PTO force [N]
%   Zmax        max stroke position [m]
%   Rmax        max PTO force slew rate [N/s]
%   predHoriz   prediction horizon [s]
%   maxIter     max MPC solver iterations
%
% Outputs:
%   MPCobj      Matlab MPC Toolbox object
%   Plant       2nd order plant used for MPC model
%   Sf          Scaling factor applied for optimization
%
% Copyright 2019 NTESS
%
% Licensed under the Apache License, Version 2.0 (the "License");
% you may not use this file except in compliance with the License.
% You may obtain a copy of the License at
%
%     http://www.apache.org/licenses/LICENSE-2.0
%
% Unless required by applicable law or agreed to in writing, software
% distributed under the License is distributed on an "AS IS" BASIS,
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
% See the License for the specific language governing permissions and
% limitations under the License.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


Ts = 0.01;
predHoriz = floor(predHoriz/Ts);

%%

exc = H_frf(:).*sqrt(Spect.S(:));

opt = tfestOptions('WeightingFilter', 1e-3+(exc/max(exc)) );
G_par = tfest(inv(frd(Zi_frf*1i*2*pi.*f, 2*pi*f)), 2,0, opt);

% figure('name','Reduced system Bode')
% bode(G_par, inv(frd(Zi_frf.*(1i*2*pi*f), 2*pi*f)))



[num2,den2] = tfdata(G_par,'v');
b0 = num2(3);
a1 = den2(2); a0 = den2(3);

sysA = [0 1;-a0 -a1];
sysB = [0;b0];
sysC = eye(2);
sysD = zeros(2,1);
SYS = ss(sysA,sysB,sysC,sysD);
Plant = c2d(SYS,Ts,'zoh');
%MPCobj=1;
%Sf=1;

%% Scaling

Sf = 20000; % Scaling factor
G = Plant*Sf;
Kp = Kp/Sf;
Ki = Ki/Sf;
Umax = Umax/Sf;
Rmax = Rmax/Sf*Ts;

%%

Kpi=[Ki Kp];
sigma=1e-5;

np=predHoriz;
nc=np;

[QMIN,RMIN,PMIN,~,~,~,~,~,~] = MPC_Cairano_YALMIP_PI_General(sigma,G.A,G.B,Kpi,np);

%%

NewPlant = G;
L = chol(PMIN);

set(NewPlant,'C',[G.C;L],'D',[G.D;zeros(2,1)], 'OutputName', '', 'OutputUnit', '' )
NewPlant.InputGroup.MV = 1;
NewPlant.OutputGroup.MO = [1 2];
NewPlant.OutputGroup.UO = [3 4];

% MPC_verbosity = mpcverbosity('off');
MPCobj = mpc(NewPlant,G.Ts,np,nc);

MPCobj.MV.Min = -Umax;
MPCobj.MV.Max = Umax;
MPCobj.MV.MinECR = 0;
MPCobj.MV.MaxECR = 0;
MPCobj.MV.RateMin = -Rmax;
MPCobj.MV.RateMax =  Rmax;

% MPCobj.MV.RateMinECR = 0.1;
% MPCobj.MV.RateMaxECR = 0.1;
MPCobj.OV(1).Min = -Zmax;
MPCobj.OV(1).Max = Zmax;
MPCobj.OV(1).MinECR = 2e-6;
MPCobj.OV(1).MaxECR = 2e-6;

ywt = sqrt(diag(QMIN))';
uwt = sqrt(diag(RMIN))';
MPCobj.Weights.OV = [ywt 0 0];
MPCobj.Weights.MV = uwt;
MPCobj.Weights.ManipulatedVariablesRate = 0.01;
MPCobj.Weights.ECR = 1e+4;
MPCobj.Model.Noise = ss(0*eye(2));

% see page 4-68 in toolbox manual (user's guide)
U = struct('Weight',uwt);
Y = struct('Weight',[0 0 1 1]);
setterminal(MPCobj,Y,U)

setoutdist(MPCobj,'model',ss(zeros(4,1)))
setEstimator(MPCobj,[],G.C);

mpcgain = dcgain(ss(MPCobj));
fprintf('\nMPC  : K = [%6.6g,%6.6g]',mpcgain(1),mpcgain(2));
fprintf('\nPI   : K = [%6.6g,%6.6g]',Ki,Kp);
fprintf('\nError: e = [%6.6g,%6.6g] (%%)\n', ...
    abs((mpcgain(1)-Ki)*100/Ki),abs((mpcgain(2)-Kp)*100/Kp));

if nargin > 7 && ~isempty(maxIter)
    MPCobj.Optimizer.UseSuboptimalSolution = true;
    MPCobj.Optimizer.MaxIter = maxIter;
end

end

function [QMIN,RMIN,PMIN,H,Fp,K_,S_,T_,Q_p] = MPC_Cairano_YALMIP_PI_General(sigma,A,B,K,np)

% This functions solves Lemma 3 in Di Cairano & Bemporad (IEEE 2010).
% To run this function, YALMIP must be installed in the computer.

nQ = size(A,1);
nR = size(B,2);
nK = size(K,2);

% Define decision variables
P = sdpvar(nQ,nQ);
Q = diag(sdpvar(nQ,1));
R = sdpvar(nR,nR);

Q_ = double2sdpvar(zeros(np*nQ,np*nQ));
K_ = zeros(np,nK);
T_ = zeros(np*nQ,nQ);
S_ = zeros(np*nQ,np);

K_(1,:) = K;
T_(1:2,1:2) = A;
S_(1:2,1) = B;
Q_(1:2,1:2) = Q;

for i = 2:np
    % K_(i,:) = K_(i-1,:)*1;
    K_(i,:) = K*(A+B*K)^(i-1);
%     if i == np
%         K_(i,:) = K*0;
%     end
    T_(2*i-1:2*i,:) = T_(2*i-3:2*i-2,:)*A;
    S_(2*i-1:2*i,1) = A*S_(2*i-3:2*i-2,1);
    for k = 2:np
        S_(2*k-1:2*k,1) = A*S_(2*k-3:2*k-2,1);
        S_(2*k-1:2*k,i) = A*S_(2*k-3:2*k-2,i);
    end
    S_(:,i) = circshift(S_(:,i-1),nQ);
    S_(1:(i-1)*nQ,i) = zeros((i-1)*nQ,1);
end

for j = 2:np
    if j < np
        Q_(1:2*j,1:2*j) = blkdiag(Q_(1:2*(j-1),1:2*(j-1)),Q);
    else
        Q_(1:2*j,1:2*j) = blkdiag(Q_(1:2*(j-1),1:2*(j-1)),P);
    end
end

% R_ = R*eye(np);
% H_ = R_+S_'*Q_*S_;
% F_ = S_'*Q_*T_;

% Define constraints
F = [P>=0,R-sigma>=0];

% Define objective function
N_ = S_'*Q_*(T_+S_*K_)+R*K_;
h = norm(N_,2);

% Solve the optimization problem
% ops = sdpsettings('solver', 'sedumi', 'sedumi.eps', 1e-8);
ops = sdpsettings('solver', 'sedumi');
optimize(F,h,ops);

% Display the solutions
QMIN = value(Q);
RMIN = value(R);
PMIN = value(P);

Q_p = zeros(np*nQ,np*nQ);
Q_p(1:2,1:2) = QMIN;

for j = 2:np
    if j < np
        Q_p(1:2*j,1:2*j) = blkdiag(Q_p(1:2*(j-1),1:2*(j-1)),QMIN);
    else
        Q_p(1:2*j,1:2*j) = blkdiag(Q_p(1:2*(j-1),1:2*(j-1)),PMIN);
    end
end

R_p = RMIN*eye(np);
H = R_p+S_'*Q_p*S_;
Fp = S_'*Q_p*T_;

end
